# License: PSF
# see: LICENSE
# for full text of the license
# 
# OO oriented Sugar Interface
# above the pysugar transport interface
#
# Original Idea: Florent Aide, <florent.aide@gmail.com>
# Design: Christophe de Vienne, <cdevienne@alphacent.com>
# Co-design : Florent Aide, <florent.aide@gmail.com>
#

from sugarobjects import SugarModule, SugarModuleCollection, SugarObject
from sugarobjects import sugar_str_field, sugar_date_field, \
        sugar_bool_field, sugar_relation_field, sugar_integer_field, \
        sugar_datetime_field, sugar_time_field, \
        init_SugarObject

class Lead(SugarObject):
    table_name = 'leads'

class User(SugarObject):
    '''
    For creating a valid user for normal usage,
    sugar needs at least these values:
        
        user.user_name = 'myloginname'
        user.last_name = 'MyLastName'
        user.user_hash = md5.new('myuserpassword').hexdigest()
        user.status = 'Active'
        user.employee_status = 'Active'
        
    '''
    table_name = 'users'

class Task(SugarObject):
    table_name = 'tasks'

class Meeting(SugarObject):
    table_name = 'meetings'

init_SugarObject(
        Lead,
        [
            sugar_str_field('portal_app'),
            sugar_str_field('portal_name'),
            sugar_bool_field('invalid_email'),
            sugar_bool_field('email_opt_out'),
            sugar_bool_field('do_not_call'),
            sugar_bool_field('deleted'),
            sugar_bool_field('converted'),
            sugar_str_field('status'),
            sugar_str_field('status_description'),
            sugar_str_field('lead_source'),
            sugar_str_field('lead_source_description'),
            sugar_str_field('description'),
            sugar_str_field('salutation'),
            sugar_str_field('title'),
            sugar_str_field('first_name'),
            sugar_str_field('last_name'),
            sugar_str_field('email1'),
            sugar_str_field('email2'),
            sugar_str_field('primary_address_street'),
            sugar_str_field('alt_address_street'),
            sugar_str_field('primary_address_city'),
            sugar_str_field('alt_address_city'),
            sugar_str_field('primary_address_postalcode'),
            sugar_str_field('alt_address_postalcode'),
            sugar_str_field('primary_address_state'),
            sugar_str_field('alt_address_state'),
            sugar_str_field('primary_address_country'),
            sugar_str_field('alt_address_country'),
            sugar_str_field('department'),
            sugar_str_field('phone_other'),
            sugar_str_field('phone_home'),
            sugar_str_field('phone_work'),
            sugar_str_field('phone_mobile'),
            sugar_str_field('phone_fax'),
            sugar_str_field('account_name'),
            sugar_str_field('account_description'),
            sugar_relation_field('account', 'account_id', 'Accounts'),
            sugar_str_field('opportunity_amount'),
            sugar_str_field('opportunity_name'),
            sugar_str_field('assigned_user_name'),
            sugar_relation_field('assigned_user', 'assigned_user_id', 'Users'),
            sugar_relation_field('reports_to', 'reports_to_id', 'Users'),
            sugar_relation_field('contact', 'contact_id', 'Contacts'),
            sugar_str_field('created_by_name'),
            sugar_relation_field('created_by', 'created_by', 'Users'),
            sugar_str_field('report_to_name'),
            sugar_str_field('modified_by_name'),
            sugar_relation_field('modified_user', 'modified_user_id', 'Users'),
            sugar_datetime_field('date_entered'),
            sugar_datetime_field('date_modified'),
            sugar_str_field('campaign_id'),
            #sugar_str_field('employee_c'),
            #sugar_str_field('Product_c'),
            sugar_str_field('refered_by'),
            #sugar_str_field('revenue_c'),
            sugar_relation_field('opportunity', 'opportunity_id',
                    'Opportunities'),
            #sugar_str_field('prospect_list_name'),
        ]
        )

init_SugarObject(
        User,
        [
            sugar_str_field('last_name'),
            sugar_str_field('employee_status'),
            sugar_str_field('user_password', send_only=True),
            sugar_str_field('phone_fax'),
            sugar_str_field('address_state'),
            sugar_str_field('messenger_id'),
            sugar_str_field('email2'),
            sugar_str_field('email1'),
            sugar_str_field('first_name'),
            sugar_datetime_field('date_entered'),
            sugar_str_field('title'),
            sugar_str_field('address_street'),
            sugar_bool_field('receive_notifications'),
            sugar_relation_field('reports_to', 'reports_to_id', 'Users'),
            sugar_relation_field('created_by', 'created_by', 'Users'),
            sugar_str_field('user_hash'),
            sugar_str_field('phone_other'),
            sugar_str_field('user_name'),
            sugar_str_field('status'),
            sugar_str_field('m_accept_status_fields'),
            sugar_str_field('description'),
            sugar_bool_field('deleted'),
            sugar_bool_field('is_group'),
            sugar_str_field('department'),
            sugar_str_field('phone_home'),
            sugar_str_field('c_accept_status_fields'),
            sugar_str_field('messenger_type'),
            sugar_str_field('address_city'),
            sugar_relation_field('modified_user', 'modified_user_id', 'Users'),
            sugar_str_field('phone_work'),
            sugar_date_field('date_modified'),
            sugar_str_field('address_country'),
            sugar_str_field('address_postalcode'),
            sugar_str_field('phone_mobile'),
            sugar_str_field('is_admin'),
            sugar_bool_field('portal_only'),
        ])
        
init_SugarObject(
        Meeting,
        [
            sugar_str_field('created_by_name'),
            sugar_date_field('date_end'),
            sugar_str_field('parent_type'),
            sugar_relation_field('assigned_user', 'assigned_user_id', 'Users'),
            sugar_integer_field('duration_minutes'),
            sugar_str_field('assigned_user_name'),
            sugar_str_field('contact_name'),
            sugar_datetime_field('date_entered'),
            sugar_str_field('outlook_id'),
            sugar_date_field('date_start'),
            sugar_relation_field('created_by', 'created_by', 'Users'),
            sugar_str_field('parent_id'),
            sugar_str_field('location'),
            sugar_str_field('modified_by_name'),
            sugar_str_field('status'),
            sugar_str_field('description'),
            sugar_bool_field('deleted'),
            sugar_time_field('time_start'),
            sugar_relation_field('modified_user', 'modified_user_id', 'Users'),
            sugar_str_field('name'),
            sugar_datetime_field('date_modified'),
            sugar_integer_field('reminder_time'),
            sugar_integer_field('duration_hours'),
        ])

init_SugarObject(
        Task,
        [
            sugar_date_field('date_due'),
            sugar_str_field('created_by_name'),
            sugar_str_field('parent_type'),
            sugar_relation_field('assigned_user', 'assigned_user_id', 'Users'),
            sugar_str_field('assigned_user_name'),
            sugar_str_field('contact_name'),
            sugar_datetime_field('date_entered'),
            sugar_str_field('date_due_flag'),
            sugar_time_field('time_due'),
            sugar_date_field('date_start'),
            sugar_relation_field('created_by', 'created_by', 'Users'),
            sugar_str_field('priority'),
            sugar_str_field('parent_id'),
            sugar_str_field('modified_by_name'),
            sugar_str_field('status'),
            sugar_str_field('description'),
            sugar_bool_field('deleted'),
            sugar_relation_field('contact', 'contact_id', 'Users'),
            sugar_time_field('time_start'),
            sugar_str_field('date_start_flag'),
            sugar_relation_field('modified_user', 'modified_user_id', 'Users'),
            sugar_str_field('name'),
            sugar_datetime_field('date_modified'),
        ])
        
class SugarStore(object):
    def __init__(self, sugar_session):
        self.backend = sugar_session
        self.m = SugarModuleCollection(self.backend)
        self.m.add('Leads', Lead)
        self.m.add('Users', User)
        self.m.add('Meetings', Meeting)
        self.m.add('Tasks', Task)

# vim: expandtab tabstop=4 shiftwidth=4:



# License: PSF
# see: LICENSE
# for full text of the license
# 
# OO oriented Sugar Interface
# above the pysugar transport interface
#
# Original Idea: Florent Aide, <florent.aide@gmail.com>
# Design: Christophe de Vienne, <cdevienne@alphacent.com>
# Co-design : Florent Aide, <florent.aide@gmail.com>

import types
import datetime
from pysugar import SugarDataError, SugarOperationnalError

DefaultBatchSize = 1000

def split_seq(seq, batchsize):
    '''
    Split a sequence into a list of batchsize long lists.
    Depending on the sequence size, last list may be
    smaller than batchsize.
    '''
    newseq = []
    elementcount = len(seq) / batchsize
    for i in xrange(0, elementcount):
        newseq.append(seq[i*batchsize:(i+1)*batchsize])
    if len(seq) % batchsize != 0:
        newseq.append(seq[elementcount*batchsize:])
    return newseq

class SugarModuleCollection:
    def __init__(self, backend):
        self.backend = backend
        self.modules = {}

    def add(self, module_name, module_class):
        self.modules[module_name] = SugarModule(
                self, module_name, module_class)
        setattr(self, module_name, self.modules[module_name])

class SugarModule(object):
    def __init__(self, collection, name, object_class):
        self.collection = collection
        self.name = name
        self.object_class = object_class
        self.elements = {}
        self.new_elements = []
        self.batch_size = DefaultBatchSize

    def add(self):
        '''
        Create a new member for the Module
        '''
        o = self.object_class(self)
        self.new_elements.append( o )
        return o

    def get(self, id):
        '''
        Fetch a member of this module by id
        '''
        if not id in self.elements.keys():
            e = self.object_class(self, id)
            self.elements[id] = e
        else:
            e = self.elements[id]
        return e

    def get_by(self, expr):
        '''
        search for the entries matching expr in this module
        '''
        entry_list = []
        table_name = self.object_class.table_name
        # XXX finish this
        raise NotImplementedError('This method is not yet implemented')

    def post(self, callback = None):
        '''
        take objects from this module in a list and post them
        with a set_entries call...
        For mass update
        callback: if specified must be a callable which accepts
        the module instance, the entry number that just finalised,
        and the lenght of the full list being posted
        to get the batch size for the module just use the returned
        module instance to access the batch_size attribute ie:

        def mycallable(module, postion, max):
            bsize = module.batch_size

        module.post(callback=mycallable)

        does not return anything
        
        '''
        element_list = []
        element_list.extend(self.new_elements)
        element_list.extend([e for e in self.elements.values()
                               if e.ismodified()])

        for i, batch in enumerate( split_seq(element_list, self.batch_size) ):
            post_list = [o.get_post_dict() for o in batch]
        
            new_ids = self.collection.backend.set_entries(self.name, post_list)
            for element, new_id in zip(batch, new_ids):
                element.finalize_post(new_id)
            if callback is not None:
                callback(self,
                    i * self.batch_size + len(batch), len(element_list))

class SugarObject(object):
    def __init__(self, module, id = None):
        self.__id = id
        self.module = module

    def get_id(self):
        return self.__id

    def isnew(self):
        return self.__id is None

    def ismodified(self):
        for prop in self.sugar_properties:
            if prop._get_modified(self):
                return True
        return False

    def get_post_dict(self):
        post_dict = {}

        if self.__id is not None:
            post_dict['id'] = self.__id

        for prop in self.sugar_properties:
            if prop._get_modified(self):
                post_dict[prop.field_name] = prop._to_sugar_value(
                        prop._get_raw_value(self))
                
        return post_dict

    def post(self):

        new_id = self.module.collection.backend.set_entry(
                self.module.name,
                self.get_post_dict()
                )

        self.finalize_post(new_id)

    def finalize_post(self, new_id):
        if self.__id is None:
            self.__id = new_id
            self.module.new_elements.pop(
                    self.module.new_elements.index(self))
            self.module.elements[new_id] = self
            self.invalidate()
        else:
            if self.__id != new_id:
                raise SugarOperationnalError(
                        'Posted object %s and received a new id: %s' % (
                                self.__id, new_id))

    def load(self):
        d = self.module.collection.backend.get_entry(
                self.module.name, self.id, '')

        for prop in self.sugar_properties:
            # respect flags for the property
            if not prop.send_only:
                prop._load_value(self, d[prop.field_name])

    def invalidate(self):
        for prop in self.sugar_properties:
            prop._cleanup(self)
        
    id = property(fget = get_id)
    
class SugarField(object):
    def __init__(self, name, field_name, module=None, read_only=False,
            receive_only=False, send_only=False, mandatory=True):
        self.name = name
        self.field_name = field_name
        self.module = module
        self.read_only = read_only
        self.receive_only = receive_only
        self.send_only = send_only
        self.mandatory = mandatory

        self.value_attr = '__sp_v_%s' % self.name
        self.modified_attr = '__sp_m_%s' % self.name

    def is_loaded(self, sugar_o):
        return hasattr(sugar_o, self.value_attr)

    def _get_raw_value(self, sugar_o):
        return getattr(sugar_o, self.value_attr)

    def __set_raw_value(self, sugar_o, value):
        setattr(sugar_o, self.value_attr, value)

    def _get_modified(self, sugar_o):
        return self.is_loaded(sugar_o) and \
            getattr(sugar_o, self.modified_attr)

    def __set_modified(self, sugar_o, value):
        setattr(sugar_o, self.modified_attr, value)

    def _cleanup(self, sugar_o):
        if hasattr(sugar_o, self.value_attr):
            delattr(sugar_o, self.value_attr)
        if hasattr(sugar_o, self.modified_attr):
            delattr(sugar_o, self.modified_attr)

    def _from_sugar_value(self, value):
        return value
    
    def _to_sugar_value(self, value):
        return value

    def _load_value(self, sugar_o, value):
        self.__set_raw_value(
                sugar_o, self._from_sugar_value(value))
        self.__set_modified(sugar_o, False)

    def _set_value(self, sugar_o, value):
        if not sugar_o.isnew() and \
                not self.is_loaded(sugar_o):
            sugar_o.load()
        self.__set_raw_value(sugar_o, value)
        self.__set_modified(sugar_o, True)

    def _get_value(self, sugar_o):
        if not sugar_o.isnew() \
                and not hasattr(sugar_o, self.value_attr):
            sugar_o.load()

        return self._get_raw_value(sugar_o)

class SugarSimpleField(SugarField):
    get_value = SugarField._get_value
    set_value = SugarField._set_value

class SugarRelationField(SugarField):
    def get_value(self, sugar_o):
        id = self._get_value(sugar_o)
        return sugar_o.module.collection.modules[
                self.module].get(id)

    def set_value(self, sugar_o, value):
        self._set_value(sugar_o, value.id)

class SugarDatetimeField(SugarField):
    '''
    sugar datetimes are represented as '2006-10-17 12:33:25'
    '''
    get_value = SugarField._get_value
    set_value = SugarField._set_value

    def _from_sugar_value(self, value):
        if not isinstance(value, types.StringType):
            raise ValueError('value should be a string')

        # TODO we should link this datetime object creation
        # with the TimeZone info from the server so we could
        # generate datetime objects with timezone set which would
        # prove really more accurate on many cases :)
        (my_date, my_time) = value.split(' ')
        (year, month, day) = my_date.split('-')
        (hour, minute, second) = my_time.split(':')
        return datetime.datetime(
                int(year), int(month), int(day),
                int(hour), int(minute), int(second))

    def _to_sugar_value(self, value):
        if not isinstance(value, datetime.datetime):
            raise ValueError('value should be a datetime.datetime')

        return value.isoformat(' ')

class SugarDateField(SugarField):
    '''
    sugar dates are stored as '2006-10-17'
    '''
    get_value = SugarField._get_value
    set_value = SugarField._set_value

    def _from_sugar_value(self, value):
        if not isinstance(value, types.StringType):
            raise ValueError('value should be a string')

        (year, month, day) = value.split('-')
        return datetime.date(int(year), int(month), int(day))

    def _to_sugar_value(self, value):
        if not isinstance(value, datetime.date):
            raise ValueError('value should be a datetime.date')

        return value.isoformat()

class SugarTimeField(SugarField):
    '''
    sugar times are stored as '12:33:25'
    '''
    get_value = SugarField._get_value
    set_value = SugarField._set_value

    def _from_sugar_value(self, value):
        if not isinstance(value, types.StringType):
            raise ValueError('value should be a string')

        (hour, minute, second) = value.split(':')
        return datetime.time(int(hour), int(minute), int(second))

    def _to_sugar_value(self, value):
        if not isinstance(value, datetime.time):
            raise ValueError('value should be a datetime.time')

        return value.isoformat()

class SugarIntegerField(SugarField):
    '''
    This field will give an integer instead of the Sugar string
    representation of an integer
    '''
    get_value = SugarField._get_value
    set_value = SugarField._set_value

    def _from_sugar_value(self, value):
        if not isinstance(value, types.StringType):
            raise ValueError('value should be a string')

        return int(value)

    def _to_sugar_value(self, value):
        if not isinstance(value, types.IntType):
            raise ValueError('value should be an integer')

        return str(value)

class SugarBooleanField(SugarField):
    '''
    when sugar stores '0/1
    we will manipulate Booleans
    '''
    get_value = SugarField._get_value
    set_value = SugarField._set_value

    def _from_sugar_value(self, value):
        if not isinstance(value, types.StringType):
            raise ValueError('value should be a string')

        try:
            v = int(value)

        except ValueError:
            # when the values are stored in different ways
            # than direct 0/1
            if value.lower() == 'off':
                return False
            elif value.lower() == 'on':
                return True

            else:
                # we must re raise the exception
                raise

        if v > 0:
            return True
        else:
            return False

    def _to_sugar_value(self, value):
        if not isinstance(value, types.BooleanType):
            raise ValueError('value should be a boolean')

        if value is True:
            return '1'
        else:
            return '0'

class SugarPropertyGetter:
    def __init__(self, sugar_field):
        self.sugar_field = sugar_field

    def __call__(self, sugar_o):
        return self.sugar_field.get_value(sugar_o)

class SugarPropertySetter:
    def __init__(self, sugar_field):
        self.sugar_field = sugar_field

    def __call__(self, sugar_o, value):
        self.sugar_field.set_value(sugar_o, value)


def create_property(sugar_field):
    return property(
            fget = SugarPropertyGetter(sugar_field),
            fset = SugarPropertySetter(sugar_field));

def init_SugarObject(sugar_object_class, fields):
    sugar_object_class.sugar_properties = []
    for f, p in fields:
        sugar_object_class.sugar_properties.append(f)
        setattr(sugar_object_class, f.name, p)

def sugar_field(name, read_only=False, receive_only=False,
        send_only=False, mandatory=True):

    field = SugarSimpleField( name, name,
            read_only=read_only,
            receive_only=receive_only, send_only=send_only,
            mandatory=mandatory,
            )

    return (field, create_property(field))

sugar_str_field = sugar_field

def sugar_relation_field(property_name, field_name, module,
        read_only=False, receive_only=False,
        send_only=False, mandatory=True):

    field = SugarRelationField(property_name, field_name,
            module=module, read_only=read_only,
            receive_only=receive_only, send_only=send_only,
            mandatory=mandatory,
            )

    return (field, create_property(field))

def sugar_datetime_field(name, read_only=False, receive_only=False,
        send_only=False, mandatory=True):

    field = SugarDatetimeField( name, name,
            read_only=read_only,
            receive_only=receive_only, send_only=send_only,
            mandatory=mandatory,
            )

    return (field, create_property(field))

def sugar_date_field(name, read_only=False, receive_only=False,
        send_only=False, mandatory=True):

    field = SugarDateField( name, name,
            read_only=read_only,
            receive_only=receive_only, send_only=send_only,
            mandatory=mandatory,
            )

    return (field, create_property(field))

def sugar_time_field(name, read_only=False, receive_only=False,
        send_only=False, mandatory=True):

    field = SugarTimeField( name, name,
            read_only=read_only,
            receive_only=receive_only, send_only=send_only,
            mandatory=mandatory,
            )

    return (field, create_property(field))

def sugar_integer_field(name, read_only=False, receive_only=False,
        send_only=False, mandatory=True):

    field = SugarIntegerField( name, name,
            read_only=read_only,
            receive_only=receive_only, send_only=send_only,
            mandatory=mandatory,
            )

    return (field, create_property(field))

def sugar_bool_field(name, read_only=False, receive_only=False,
        send_only=False, mandatory=True):

    field = SugarBooleanField( name, name,
            read_only=read_only,
            receive_only=receive_only, send_only=send_only,
            mandatory=mandatory,
            )

    return (field, create_property(field))

# vim: expandtab tabstop=4 shiftwidth=4:



#!/usr/bin/env python
# License: PSF
# see: LICENSE
# for full text of the license
#
major_version = 0
mid_version = 0
minor_version = 6
version='%s.%s.%s' % (major_version, mid_version, minor_version)



#!/usr/bin/env python
import ez_setup
ez_setup.use_setuptools()
from setuptools import setup, find_packages
import os

pysugar_dir = 'pysugar'
version_file = 'pysugar_version.py'
lines = file(os.path.join(pysugar_dir, version_file), 'r').readlines()

# run the content of version file
for line in lines:
    exec(line)

setup(
    name = "pysugar",
    version = version,
    packages = find_packages(),
    #scripts = [''],
    zip_safe = False,

    # Project uses pytz
    install_requires = ['pytz>=2006j',
        'elementsoap>=0.5', 'elementtree>=1.2a5' ],
    # no more specific dependency
    dependency_links = [],

    package_data = {
        # If any package contains *.txt or *.rst files, include them:
        '': ['*.txt', '*.php'],
        # And include any *.msg files found in the 'hello' package, too:
        #'hello': ['*.msg'],
    },

    # metadata for upload to PyPI
    author = "Florent Aide, Christophe de Vienne",
    author_email = "florent.aide@gmail.com, cdevienne@alphacent.com",
    description='Python binding of SugarCRM, \
that uses the SOAP API provided by SugarCRM',

    long_description='''A pythonic binding for the SugarCRM SOAP interface,
    make it possible to use objects as if they were local ie:

    >>> import pysugar
    >>> sugar_user = 'myuser'
    >>> sugar_password = 'mypassword'
    >>> sugar_base_url = 'http://myserver/sugar'
    >>> sugar_debug = False
    >>> sugar_session = pysugar.SugarSession(
    ... sugar_user, sugar_password, sugar_base_url, sugar_debug)
    >>> sugar_store = pysugar.SugarStore(sugar_session)
    >>> sugar_lead = sugar_store.m.Leads.add()
    >>> sugar_lead.first_name = 'Test Lead'
    >>> sugar_lead.post()
    >>> print sugar_lead.id
    929a26ac-fc47-3232-20a6-4534cdb3290e

    then you can test the result in another session

    ...(initialization stuff)
    >>> id = '929a26ac-fc47-3232-20a6-4534cdb3290e'
    >>> sugar_lead = sugar_store.m.Leads.get(id)
    >>> print sugar_lead.first_name

    id being the previously obtained id string from the sugar server
    ''',
    license = "PSF",
    keywords = "Sugar CRM SOAP",
    url = "",   # project home page, if any

    # could also include download_url, classifiers, etc.
    classifiers = """\
    Development Status :: 3 - Alpha
    Intended Audience :: Developers
    License :: OSI Approved :: Python Software Foundation License
    Programming Language :: Python
    Topic :: Software Development :: Libraries :: Python Modules
    Operating System :: OS Independent
    """
)


# vim: expandtab tabstop=4 shiftwidth=4:



#!/usr/bin/env python
# License: PSF
# see: LICENSE
# for full text of the license
#
from elementsoap import ElementSOAP
from elementtree.ElementTree import tostring, dump
import md5
import xml
from pytz import timezone
import datetime
import base64
import urllib2
import types
from pysugar_version import version, major_version, minor_version, mid_version

__version__ = version
__author__ = "Florent Aide"
__doc__ = '''This is the transport layer for the pysugar library.
It uses ElementSOAP in order to communicate with the Sugar NuSoap Server.
'''

def item_to_name_value(item):
    '''
    takes a real item dictionnary and prepare it for soap
    returns a name value formated list of dictionnaries
    '''
    nv_list = []
    for key in item:
        nv_list.append({'name':key, 'value':item[key]})
    
    return nv_list

def name_value_to_item(nv_list):
    '''
    takes a name/value list from a sugar answer and returns a
    dictionnary mapping item
    '''
    #print tostring(nv_list)
    item = {}

    #piece = nv_list.find('item')
    item['id'] = nv_list.findtext('id') 
    item['module_name'] = nv_list.findtext('module_name')

    nv = nv_list.find('name_value_list')
    for el in nv.findall('item'):
        item[el.findtext('name')] = el.findtext('value')

    return item
        
class SugarError(Exception):
    '''
    This is the base class for our errors
    '''
    pass

class SugarOperationnalError(SugarError):
    '''
    when some unknow & weird error occurs
    '''
    pass

class SugarLoginError(SugarError):
    '''
    When a login error occurs, we raise this one
    '''
    pass

class SugarLogoutError(SugarError):
    '''
    in case of error during logout phase
    '''
    pass

class SugarCredentialError(SugarError):
    '''
    When a function call requiring a valid session id is call without
    one, this exception will be raised
    '''
    pass

class SugarVersionError(SugarError):
    '''
    If an error occurs on the version number
    '''
    pass

class SugarConnectError(SugarError):
    '''
    any connection error
    '''
    pass

class SugarDataError(SugarError):
    '''
    An error where the data from the server does not macth what we
    were expecting.
    '''
    pass

class SugarService(ElementSOAP.SoapService):
    '''
    This is the transport part of pysugar, it implements the soap
    calls necessary to talk to the NuSOAP server.
    This class should no be used directly by end-users.
    End-users should prefer to use the SugarSession class
    that encapsulates the notion of session and takes care of
    providing the session_id to the underlying sugar service
    '''
    def __init__(self, url):
        self.url = url
        self.application_name = "pysugar"
        ElementSOAP.SoapService.__init__(self, url)
    
    def login(self, user, password):
        """
        this will log into sugar given a username and a password
        it will return the uuid for the session

        if an error occurs, a SugarLoginError is raised.
        """
        pass_hash = md5.new(password).hexdigest()
        action = 'login'
        request = ElementSOAP.SoapRequest(action)
        user_auth = ElementSOAP.SoapElement(request, "user_auth")
        ElementSOAP.SoapElement(user_auth, "user_name", "string", user)
        ElementSOAP.SoapElement(user_auth, "password", "string", pass_hash)
        ElementSOAP.SoapElement(user_auth, "version", "string", '1.2')
        ElementSOAP.SoapElement(request, "application_name",
                "string", self.application_name)
        
        response = self.call(action, request)
        result = response.find('return')
        error_element = result.find('error')
        if not error_element.findtext('number') == '0':
            # there was a problem we need to raise an exception
            raise SugarLoginError('number: %s, name: %s, desc: %s' % (
                    error_element.findtext('number'),
                    error_element.findtext('name'),
                    error_element.findtext('description')
                    ))

        return result.findtext('id')
    
    def logout(self, session_id):
        '''
        given a session id, will make sure to invalidate this session
        so it cannot be used subsequently.
        For security reasons it is important to use this method
        '''
        action  = 'logout'
        request = ElementSOAP.SoapRequest(action)
        ElementSOAP.SoapElement(request, "session", "string", session_id)

        response = self.call(action, request)
        ret = response.find('return')
        error = int(ret.findtext('number'))
        if error:
            name = ret.findtext('name')
            desc = ret.findtext('description')
            raise SugarLogoutError('name: "%s", desc: "%s"' % (name, desc))

    def get_user_id(self, session_id):
        '''
        given a session_id, will return the user_id that is logged in
        '''
        action = 'get_user_id'
        request = ElementSOAP.SoapRequest(action)
        ElementSOAP.SoapElement(request, "session", "string", session_id)

        response = self.call(action, request)
        return response.findtext('return')

    def test(self, teststring):
        '''
        a ping function to make sure the NuSOAP server is up and running
        '''
        action = 'test'
        request = ElementSOAP.SoapRequest(action)
        ElementSOAP.SoapElement(request, "string", "string", teststring)
        
        response = self.call(action, request)
        return response.findtext('return')

    def get_gmt_time(self):
        '''
        '''
        action = 'get_gmt_time'
        request = ElementSOAP.SoapRequest(action)

        response = self.call(action, request)
        return response.findtext('return')

    def get_server_version(self):
        '''
        '''
        action = 'get_server_version'
        request = ElementSOAP.SoapRequest(action)

        response = self.call(action, request)
        return response.findtext('return')

    def set_entry(self, session_id, module, item):
        '''
        creates an item in the given module
        module is a string representing the module ie: 'Leads' or 'Contacts'
        This will return the id for the newly created item or raise
        a SugarError in case of problem.
        '''
        action = 'set_entry'
        request = ElementSOAP.SoapRequest(action)
        ElementSOAP.SoapElement(request, "session", "string", session_id)
        ElementSOAP.SoapElement(request, "module", "string", module)
        nv_list = ElementSOAP.SoapElement(request, "name_value_list", "Array")
        for key in item:
            item_el = ElementSOAP.SoapElement(nv_list, "item")
            ElementSOAP.SoapElement(item_el, 'name', 'string', key)
            ElementSOAP.SoapElement(item_el, 'value', 'string', item[key])

        #print tostring(request)

        response = self.call(action, request)
        ret = response.find('return')

        error_elem = ret.find('error')
        error = int(error_elem.findtext('number'))

        if error:
            name = error_elem.findtext('name')
            desc = error_elem.findtext('description')
            raise SugarError('number: %s, name: "%s", desc: "%s"' % (
                    error, name, desc))

        return ret.findtext('id')
    
    def set_entries(self, session_id, module, items):
        '''
        create multiple entries at the same time in the specified module
        '''
        action = 'set_entries'
        request = ElementSOAP.SoapRequest(action)
        ElementSOAP.SoapElement(request, "session", "string", session_id)
        ElementSOAP.SoapElement(request, "module", "string", module)
        vlists = ElementSOAP.SoapElement(request, "name_value_lists", "Array")
        for item in items:
            vlist = ElementSOAP.SoapElement(vlists, "name_value_list", "Array")
            for key in item:
                item_el = ElementSOAP.SoapElement(vlist, "item")
                ElementSOAP.SoapElement(item_el, 'name', 'string', key)
                ElementSOAP.SoapElement(item_el, 'value', 'string', item[key])

        #print tostring(request)
        
        response = self.call(action, request)
        ret = response.find('return')

        error_elem = ret.find('error')
        error = int(error_elem.findtext('number'))

        if error:
            name = error_elem.findtext('name')
            desc = error_elem.findtext('description')
            raise SugarError('number: %s, name: "%s", desc: "%s"' % (
                    error, name, desc))

        ids_el = ret.find('ids').findall('item')
        ids = []
        for el in ids_el:
            ids.append(el.text)

        return ids

    def get_entry(self, session_id, module, id, select_fields):
        '''
        given a session_id this method enables us
        to interrogate a specified module
        an to ask for a specific object by id
        the select_fields is a way to ask for only
        some fields to be returned instead
        of the whole object : currently does nothing...

        In case of error this module will raise a SugarError exception
        '''
        action = 'get_entry'
        request = ElementSOAP.SoapRequest(action)
        ElementSOAP.SoapElement(request, "session", "string", session_id)
        ElementSOAP.SoapElement(request, "module", "string", module)
        ElementSOAP.SoapElement(request, "id", "string", id)
        ElementSOAP.SoapElement(request, "selection", "list", select_fields)

        response = self.call(action, request)
        ret = response.find('return')
        entry_list = ret.find('entry_list')

        error_elem = ret.find('error')
        error = int(error_elem.findtext('number'))

        if error:
            name = error_elem.findtext('name')
            desc = error_elem.findtext('description')
            raise SugarError('number: %s, name: "%s", desc: "%s"' % (
                    error, name, desc))

        item = entry_list.find('item')

        return name_value_to_item(item)

    def get_entry_list(self, session_id, module, query, order_by,
                offset, selection, max_result, deleted):
        '''
        Get a list of entries for a specified module in essence the same
        as a get_entry call where the result will be a list of items
        instead of just one.
        query must be of the form:
        "leads.last_name LIKE 'T%'"
        or
        "leads.last_name is not NULL"
        '''
        action = 'get_entry_list'
        request = ElementSOAP.SoapRequest(action)
        ElementSOAP.SoapElement(request, "session", "string", session_id)
        ElementSOAP.SoapElement(request, "module", "string", module)
        ElementSOAP.SoapElement(request, "query", "string", query)
        ElementSOAP.SoapElement(request, "order_by", "string", order_by)
        ElementSOAP.SoapElement(request, "offset", "integer", offset)
        ElementSOAP.SoapElement(request, "select_fields", "string", selection)
        ElementSOAP.SoapElement(request, "max_results", "integer", max_result)
        ElementSOAP.SoapElement(request, "deleted", "integer", deleted)

        response = self.call(action, request)
        ret = response.find('return')
        error_elem = ret.find('error')
        error = int(error_elem.findtext('number'))

        if error:
            name = error_elem.findtext('name')
            desc = error_elem.findtext('description')
            raise SugarError('number: %s, name: "%s", desc: "%s"' % (
                    error, name, desc))

        entries = ret.find('entry_list').findall('item')
        elist = []
        for entry in entries:
            elist.append(name_value_to_item(entry))

        # TODO: we should also return the next offset and friends...
        return elist

    def get_available_modules(self, session_id):
        '''
        returns the list of modules names (strings)
        we can access with our login
        '''
        action = 'get_available_modules'
        request = ElementSOAP.SoapRequest(action)
        ElementSOAP.SoapElement(request, "session", "string", session_id)

        response = self.call(action, request)
        ret = response.find('return')
        error_elem = ret.find('error')
        error = int(error_elem.findtext('number'))

        if error:
            name = error_elem.findtext('name')
            desc = error_elem.findtext('description')
            raise SugarError('number: %s, name: "%s", desc: "%s"' % (
                    error, name, desc))

        modules_el = ret.find('modules')
        modules_list = []
        for i in modules_el:
            modules_list.append(i.text)

        return modules_list

    def get_note_attachment(self, session_id, attachement_id):
        '''
        given a session_id and an attachement_id this function will
        return a 2-tuple containing the filename and the actual file content
        the file content is binary data as if it were the result of a read()
        operation on a file like object...
        '''
        action = 'get_note_attachment'
        request = ElementSOAP.SoapRequest(action)
        ElementSOAP.SoapElement(request, "session", "string", session_id)
        ElementSOAP.SoapElement(request, "id", "string", attachement_id)

        response = self.call(action, request)
        ret = response.find('return')
        error_elem = ret.find('error')
        error = int(error_elem.findtext('number'))

        if error:
            name = error_elem.findtext('name')
            desc = error_elem.findtext('description')
            raise SugarError('number: %s, name: "%s", desc: "%s"' % (
                    error, name, desc))

        note_attachm = ret.find('note_attachment')
        filename = note_attachm.findtext('filename')
        file = base64.b64decode(note_attachm.findtext('file'))

        return (filename, file)

    def get_relationships(self, session_id, module_name, module_id,
            related_module, related_module_query, deleted=False):
        '''
        session_id: a valid session id
        module_name: a module name such as 'Accounts'
        module_id: a valid object id in the module name
        related_module: a valid module name such as 'Contacts' that
                may have objects related the the already given module_id
        related_module_query: a string representing a subset of the
                SQL Query that will be executed on the server!!!
                (seems dangerous)
        deleted: a boolean. If True, then all records will be returned even
                the deleted ones. Defaults to False if nothing given


        returns: a list of ids for the related module objects that were found
                matching your search criterions


        Discussion:

        There are some limitations currently enforced on the server side
        by the PHP code. You have only access to a limited amount of base
        modules (module_name) and for each module_name you only can
        name a restricted list of related modules...

        If you specify something else the server will return an error stating
        that the module does not exist

        An excerpt of the
        PHP comment follows:

        ----
        * Only the listed combinations below are supported.  On the left is the
        * primary module.  Under each primary module is a
        * list of available related modules.
        *  'Contacts'=>
        *              'Calls'
        *              'Meetings'
        *              'Users'
        *  'Users'=>
        *              'Calls'
        *              'Meetings'
        *              'Contacts'
        *  'Meetings'=>
        *              'Contacts'
        *              'Users'
        *  'Calls'=>
        *              'Contacts'
        *              'Users'
        *  'Accounts'=>
        *              'Contacts'
        *              'Users'
        ----

        Exemple Use:
        >>> rels = sugar_session.get_relationships(
        '644b1fa1de5937ba1484516777e34aae',
        'Accounts', 'e8151a7f-136d-3b8e-b026-456ae620866d',
        'Contacts', '')
        >>> rels
        ['4841ac25-083a-f629-da32-456b061467d0']

        Background:
        the PHP code executed on the server is this one:

        ----
        $result = $r_mod->db->query("SELECT id, date_modified FROM " \
        .$r_mod->table_name \
        . " WHERE id IN $in AND ( $related_module_query )");
        ----

        as you can see: the related module query is passed
        'as is' in the WHERE clause.
        Please note we are NOT the authors of the PHP code. For any inquiries
        about this you should direct your questions to the Sugar CRM team.

        '''
        if not isinstance(deleted, types.BooleanType):
            msg='deleted keyword must be an integer, not: %s' % type(deleted)
            raise ValueError(msg)

        action = 'get_relationships'
        request = ElementSOAP.SoapRequest(action)
        ElementSOAP.SoapElement(request, "session", "string", session_id)
        ElementSOAP.SoapElement(request, "module_name", "string", module_name)
        ElementSOAP.SoapElement(request, "module_id", "string", module_id)
        ElementSOAP.SoapElement(request, "related_module",
                "string", related_module)
        ElementSOAP.SoapElement(request, "related_module_query",
                "string", related_module_query)
        ElementSOAP.SoapElement(request, "deleted", "string", deleted)

        response = self.call(action, request)
        ret = response.find('return')
        error_elem = ret.find('error')
        error = int(error_elem.findtext('number'))

        if error:
            name = error_elem.findtext('name')
            desc = error_elem.findtext('description')
            raise SugarError('number: %s, name: "%s", desc: "%s"' % (
                    error, name, desc))

        items_el = ret.find('ids')
        id_list = []
        for item in items_el:
            id_list.append(item.findtext('id'))

        return id_list

    def set_relationship(self, session_id, module1,
            module1_id, module2, module2_id):
        '''
        set a relationship beetween module1_id and module2_id
        the same rules apply that are described in get_relationships
        '''
        action = 'set_relationships'
        request = ElementSOAP.SoapRequest(action)
        ElementSOAP.SoapElement(request, "session", "string", session_id)
        srv = ElementSOAP.SoapElement(request, "set_relationship_value")
        ElementSOAP.SoapElement(srv, "module1", "string", module1)
        ElementSOAP.SoapElement(srv, "module1_id", "string", module1_id)
        ElementSOAP.SoapElement(srv, "module2", "string", module2)
        ElementSOAP.SoapElement(srv, "module2_id", "string", module2_id)

        response = self.call(action, request)
        ret = response.find('return')
        error_elem = ret.find('error')
        error = int(error_elem.findtext('number'))

        if error:
            name = error_elem.findtext('name')
            desc = error_elem.findtext('description')
            raise SugarError('number: %s, name: "%s", desc: "%s"' % (
                    error, name, desc))

        return ret

    def prune_meetings(self, session_id, date_from, date_to):
        '''
        remove ALL the meetings from the database !!!
        USE AT YOUR OWN RISK.
        You need to know what you want to do when using this function
        please make sure you do some tests before using this function
        on a live server
        This function is ONLY available if you install the meeting extension
        from the install directory of the pysugar distribution
        '''
        assert isinstance(date_from, datetime.date), \
            "date_from should be a datetime.date instance"
        assert isinstance(date_to, datetime.date), \
            "date_to should be a datetime.date instance"

        action = 'prune_meetings'
        request = ElementSOAP.SoapRequest(action)
        ElementSOAP.SoapElement(request, "session", "string", session_id)
        ElementSOAP.SoapElement(request, "date_from",
            "string", date_from.strftime('%Y-%m-%d'))
        ElementSOAP.SoapElement(request, "date_to", 
            "string", date_to.strftime('%Y-%m-%d'))
        response = self.call(action, request)

        ret = response.find('return')
        error_elem = ret.find('error')
        error = int(error_elem.findtext('number'))

        if error:
            name = error_elem.findtext('name')
            desc = error_elem.findtext('description')
            raise SugarError('number: %s, name: "%s", desc: "%s"' % (
                    error, name, desc))

        return ret

class SugarSession:
    '''
    This class is the entry point to the rest of the API.
    A Sugar Session is the first thing we need in order to
    communicate with the Sugar Server.
    When we have a session all other requests to the server will
    be called on the session object.
    '''
    def __init__(self, username, password, base_url,
            debug=False, user_management=False, nusoapfile='soap.php'):
        '''
        username: a string representing the login
        password: a string with the password for the login
        base_url: a string containing the url for the sugar server
                without the soap.php?wsdl trailer
        debug: a boolean, False by default
        user_management: a boolean, False by default
        nusoapfile: the name of the nusoap file (php script) that will
        be used. This is here so that users can write their own php nusoap
        servers for sugar and connect to it.
        
        example:
            s = SugarSession('myuser', 'mypass', 'http://myserver/sugar')
        
        Here s should be a pysugar.SugarSession object ready to be used
        '''
        self.user_management = user_management
        self._session_id = False
        self._debug = debug

        self.base_url = base_url
        self.application_name = 'pysugar'
        #soap_url = base_url + '/soap.php?wsdl'
        #user_url = base_url + '/soap_users.php?wsdl'
        soap_url = base_url + "/" + nusoapfile
        
        try:
            urllib2.urlopen(soap_url)
        except urllib2.HTTPError, e:
            msg = "Can't resolve %s." % e.geturl()
            raise SugarConnectError(msg)

        if user_management:
            try:
                urllib2.urlopen(user_url)
            except urllib2.HTTPError, e:
                msg = "Can't resolve %s.\n" % e.geturl()
                msg += "Maybe you should deploy the soap_users.php script ?"
                raise SugarConnectError(msg)

        self.service = SugarService(soap_url)
        #try:
        #    self.soap_proxy = SOAPpy.WSDL.Proxy(soap_url)
        #
        #except xml.parsers.expat.ExpatError, args:
        #    msg = 'Cannot communicate with server,'
        #    msg += 'please check the url. msg= %s' % args
        #    raise SugarError(msg)

        #if user_management:
        #    try:
        #        self.user_proxy = SOAPpy.WSDL.Proxy(user_url)
        #    except xml.parsers.expat.ExpatError, args:
        #        msg = 'Cannot communicate with server,'
        #        msg += 'please check the url. msg= %s' % args
        #        raise SugarError(msg)

        if not self.test():
            msg = "The Sugar server seems to be offline"
            msg += "or unresponsive"
            raise SugarError(msg)
        else:
            self.login(username, password)


    def __validate_login(self):
        '''
        this method is for internal use only. It is used to make sure
        there as been a successful login in the past for our session.
        Each method that requires a valid session should call this before
        anything else.
        If no session has been opened already this method will spit an
        SugarCredentialError exception
        If session is OK this method returns nothing
        '''
        if not self._session_id:
            raise SugarCredentialError(
                    'A valid session id is required to use this method')

    def login(self, username, password):
        '''
        tries to log into the Sugar server with the given credentials
        '''
        session_id = self.service.login(username, password)
        self._session_id = session_id

        if self._debug:
            print "We are logged-in with user-id: '%s'" %\
            self.get_user_id()
            print "Our session id is: '%s'" % self._session_id

    def get_entry_list(self, module, query, order_by,
            offset, selection, max_result, deleted):
        '''
        this raw call enables to ask any modules to give its list of data
        ie: Leads module will return a list of leads selected according to
        our search criterions.
        the query string must be of the form:
        "leads.last_name LIKE 'T%'"
        this is basically an SQL WHERE clause. To have more abstraction about
        these things we will use the sugarobjects module.
        '''
        self.__validate_login()

        return self.service.get_entry_list(self._session_id,
                module, query, order_by, offset,
                selection, max_result, deleted)
    
        #next_offset = res.next_offset
        #field_list = res.field_list
        #module_list = res.entry_list
        #errors = res.error
        #num_result = res.result_count
    
        #my_items = []

        #for item in module_list:
        #    my_items.append(dict(item.name_value_list))

        #return my_items

    def get_entry(self, module, id, selection):
        '''
        This method is the way to get entries according to their ids
        '''
        self.__validate_login()
        return self.service.get_entry(self._session_id, module,
                id, selection)

    def set_entry(self, module, item):
        '''
        create a new entry in Sugar
        if an error occurs a SugarError will be raised
        if the operation is a success, a string representing the id of the
        created/modified entry will be returned
        '''
        self.__validate_login()
        return self.service.set_entry(self._session_id, module, item)

    def set_entries(self, module, items):
        '''
        create a batch of new entries in Sugar for the same module
        '''
        self.__validate_login()
        return self.service.set_entries( self._session_id,
                module, items)

    def set_note_attachment(self, note):
        '''
        create a new note attachment in Sugar
        '''
        raise NotImplementedError
    
    def get_note_attachment(self, id):
        '''
        retrieve a note attachement from sugar using its id
        returns a 2-tuple containing the filename of the attachment
        and the binary content of the file attached to the note
        '''
        self.__validate_login()
        res = self.service.get_note_attachment(self._session_id, id)

        return res
    
    def relate_note_to_module(self, note_id, module_name, module_id):
        '''
        ???
        '''
        raise NotImplementedError

    def get_related_notes(self, module_name, module_id, selection):
        '''
        ???
        '''
        raise NotImplementedError

    def logout(self):
        '''
        destroy the given session on the server
        '''
        self.__validate_login()
        res = self.service.logout(self._session_id)
        # make sure session id is now invalid
        self._session_id = None
    
    def get_user_id(self):
        '''
        this will return a user id string
        if the session is invalid the returned user id will be '-1'
        '''
        self.__validate_login()
        return self.service.get_user_id(self._session_id)

    def get_module_fields(self, module_name):
        self.__validate_login()
        return self.soap_proxy.get_module_fields(
                self._session_id,
                module_name
                )
                
    def get_available_modules(self):
        '''
        Lists the modules present on the server we are logged in
        '''
        self.__validate_login()
        res = self.service.get_available_modules(self._session_id)

        return res
        #error = res.error
        #if error.number == '0':
        #    modules = res.modules
        #    return list(modules)
        #else:
        #    raise SugarError('%s' % error.description)

    def update_portal_user(self, username, name_value_list):
        '''
        ???
        '''
        raise NotImplementedError
    
    def create_user(self, user_name, password):
        if not self.user_management:
            raise SugarError('User Management is not initialized')

        self.__validate_login()

        res = self.user_proxy.create_user(
                self._session_id,
                user_name,
                password)

        if res.id == '-1':
            print res.error
        return res.id

    def test(self):
        '''
        test that the given url responds something we can use
        and the NuSoap php xmlrpc server is indeed present
        '''
        onetest = 'OneStringOfTest'
        #res = self.soap_proxy.test(onetest)
        res = self.service.test(onetest)
        if res == onetest:
            return True
        else:
            return False

    def get_server_time(self):
        '''
        returns the time on the server
        in local time.
        I did not implement this one since we have the GMT one
        and times seem to alway come back in UTC from sugar server:
        ie: when you look at a last modified timestamp from a sugar lead
        you will see that it is sent to you in UTC time by the server.
        '''
        raise NotImplementedError
    
    def get_gmt_time(self):
        '''
        returns the server time in GTM
        the returned time is a datetime.datetime instance in UTC timezone
        the timezone IS specified in the intance using the pytz library
        You can use this function without being logged into the server.
        '''
        res = self.service.get_gmt_time()
        utc = timezone('UTC')

        (ymd, hms) = res.split(' ')

        (year, month, day) = ymd.split('-')
        (hour, minute, second) = hms.split(':')

        year = int(year)
        month = int(month)
        day = int(day)
        hour = int(hour)
        minute = int(minute)
        second = int(second)
        
        return datetime.datetime(
                year,
                month,
                day,
                hour,
                minute,
                second,
                0,
                tzinfo=utc
                )
    
    def get_server_version(self):
        '''
        returns the version number of the server
        this function can be used without a login...
        '''
        res = self.service.get_server_version()
        if res == '1.0':
            msg = 'The server does not disclose this information'
            raise SugarVersionError(msg)
        else:
            return res

    def get_user_role_ids(self, user_id):
        if not self.user_management:
            raise SugarError('User Management is not initialized')

        self.__validate_login()

        return self.user_proxy.get_user_role_ids(
                self._session_id,
                user_id)

    def remove_user_role(self, role_id, user_id):
        if not self.user_management:
            raise SugarError('User Management is not initialized')

        self.__validate_login()

        return self.user_proxy.remove_user_role(
                self._session_id,
                role_id,
                user_id)

    def get_relationships(self, module_name, module_id, related_module,
            related_module_query, deleted=False):
        '''
        session_id: a valid session id
        module_name: a module name such as 'Accounts'
        module_id: a valid object id in the module name
        related_module: a valid module name such as 'Contacts' that
                may have objects related the the already given module_id
        related_module_query: a string representing a subset of the
                SQL Query that will be executed on the server!!!
                (seems dangerous)
        deleted: a boolean. If True, then all records will be returned even
                the deleted ones. Defaults to False if nothing given


        returns: a list of ids for the related module objects that were found
                matching your search criterions


        Discussion:

        There are some limitations currently enforced on the server side
        by the PHP code. You have only access to a limited amount of base
        modules (module_name) and for each module_name you only can
        name a restricted list of related modules...

        If you specify something else the server will return an error stating
        that the module does not exist

        An excerpt of the
        PHP comment follows:

        ----
        * Only the listed combinations below are supported.  On the left is the
        * primary module.  Under each primary module is a list of
        * available related modules.
        *  'Contacts'=>
        *              'Calls'
        *              'Meetings'
        *              'Users'
        *  'Users'=>
        *              'Calls'
        *              'Meetings'
        *              'Contacts'
        *  'Meetings'=>
        *              'Contacts'
        *              'Users'
        *  'Calls'=>
        *              'Contacts'
        *              'Users'
        *  'Accounts'=>
        *              'Contacts'
        *              'Users'
        ----

        Exemple Use:
        >>> rels = sugar_session.get_relationships(
        '644b1fa1de5937ba1484516777e34aae',
        'Accounts', 'e8151a7f-136d-3b8e-b026-456ae620866d',
        'Contacts', '')
        >>> rels
        ['4841ac25-083a-f629-da32-456b061467d0']

        Background:
        the PHP code executed on the server is this one:

        ----
        $result = $r_mod->db->query("SELECT id, date_modified FROM " \
        .$r_mod->table_name \
        . " WHERE id IN $in AND ( $related_module_query )");
        ----

        as you can see: the related module query is passed
        'as is' in the WHERE clause.
        Please note we are NOT the authors of the PHP code. For any inquiries
        about this you should direct your questions to the Sugar CRM team.
        '''
        self.__validate_login()
        return self.service.get_relationships(
               self._session_id,
               module_name,
               module_id,
               related_module,
               related_module_query,
               deleted)
        

    def set_relationship(self, module1, module1_id, module2, module2_id):
        self.__validate_login()
        return self.service.set_relationship(
                self._session_id, module1, module1_id,
                module2, module2_id)
            
    def set_relationships(self, set_relationship_list):
        '''
        ???
        '''
        raise NotImplementedError
    
    def set_document_revision(self, document_revision):
        '''
        ???
        '''
        raise NotImplementedError
    
    def prune_meetings(self, date_from, date_to):
        '''
        this method will only work if the meeting extension has been installed
        on the server. All information regarding this extension can be found
        in the install directory or the pysugar distribution.
        '''
        self.__validate_login()
        return self.service.prune_meetings(
            self._session_id, date_from, date_to)

# vim: expandtab tabstop=4 shiftwidth=4:
